import { AdapterManager as AdapterManagerInterface, TransportService, TransportEvent, Transport } from "../interface/connector.interface"
import { TransmitterAdapter } from './adapter.transmitter'
import { ReceiverAdapter } from './adapter.receiver'
import { v4 as uuidv4 } from 'uuid'
import { Subject } from "rxjs"
import { WebsocketTransportService } from "../transport/websocket"
import { HttpTransportService } from "../transport/http"
import config from '../config/config.json';
import ConsoleLogger from "../utils/log.utils"
import { RequestResponseAdapter } from "./adapter.request.response"
import { AdapterSet } from "../interface/general.interface"

export class AdapterManager implements AdapterManagerInterface {
    private console: ConsoleLogger = new ConsoleLogger(`AdapterManager`, ['managers'])
    private transportServiceArray: TransportService[] = []
    private transportSet: TransportSet[] = []
    private adapterSet: AdapterSet[] = []
    private event!: Subject<TransportEvent>

    constructor(event: Subject<TransportEvent>, browserEnv?: boolean | undefined) {
        this.event = event
        this.console.log({ message: `Contructing self...` })

        this.sortTransportFromEnv(this.transportSet)
        this.transportSet.forEach(set => {
            this.setUpTransportService(set, event, browserEnv)
        })
    }

    /* This one change to subscribe since I want it to return an array of adapters instead of just a set of them, becauise there could
    mulitple adapters that the manager will have to work with */
    public getAdapter(clientId: string): AdapterSet | null {
        this.console.log({ message: `Instantiating an adapter set....` })
        let transportType: Transport = process.env.Transport as unknown as Transport // as default  for now
        let adapterId: string = clientId
        let transportService: TransportService | undefined = this.transportServiceArray.find(obj => obj.getInfo() == transportType)
        if (transportService) {
            let transmitterAdapter: TransmitterAdapter = new TransmitterAdapter(adapterId, transportType, transportService)
            let receiverAdapter: ReceiverAdapter = new ReceiverAdapter(adapterId, transportType, transportService)
            let adapterSet: AdapterSet = {
                id: adapterId,
                dateCreated: new Date(),
                transmitterAdapter: transmitterAdapter,
                receiverAdapter: receiverAdapter,
                requestResponsAdapter: new RequestResponseAdapter(transmitterAdapter, receiverAdapter)
            }
            this.adapterSet.push(adapterSet)

            this.event.next({
                id: uuidv4(),
                event: 'New Adapter',
                data: adapterId
            } as TransportEvent)
            return adapterSet
        } else {
            return null
        }
    }

    // Server to be set up as well as acquiring client information if needed. Like in the case for grpc and socket. Http not requ`ired.
    private setUpTransportService(transportSet: TransportSet, event: Subject<TransportEvent>, isClient?: boolean): void {
        try {
            let transportService: TransportService = this.instantiateTransportService(transportSet.transport, event)
            this.transportServiceArray.push(transportService)
            if (transportService instanceof WebsocketTransportService) {
                this.console.log({ message: `Just Double Checking... this is websocket` })
                if (isClient) {
                    // please note this is subject to change depending on the UI environemnt. Angular has their own built in function to read json file based on Swopt-UI
                    transportService.startClient(config.connection.transmitter)
                } else {
                    transportService.startServer(transportSet.port);
                }
            } else if (transportService instanceof HttpTransportService) {
                this.console.log({ message: `Just Double Checking... this is http` })
                // Additional Http-specific setup if needed.
                if (isClient) {
                    transportService.startClient(`http://localhost:3000`)
                } else {
                    transportService.startServer(transportSet.port)
                }
            }
        } catch (error) {
            this.console.error({ message: 'Fail to set transport. Error in setting up transport', details: error })
        }
    }

    private instantiateTransportService(transportType: Transport, event: Subject<TransportEvent>): TransportService {
        if (transportType == Transport.Websocket) {
            return new WebsocketTransportService(event)
        }
        else if (transportType == Transport.Http) {
            return new HttpTransportService(event)
        } else {
            throw new Error(`No Transport Service Instantiated`)
        }
    }

    private sortTransportFromEnv(transportSet: TransportSet[]): void {
        let transportList: string[] = process.env.Transport?.split(',') || []
        let portList: number[] = (process.env.PORT?.split(',') || []).map(port => Number(port));
        transportList.forEach((transport, index) => {
            transportSet.push({ transport: transport, port: portList[index] } as unknown as TransportSet)
        })
        this.console.log({ message: 'TransportSetList', details: this.transportSet })
    }

}


interface TransportSet {
    transport: Transport,
    port: number
}

