import { Socket as ClientSocket, io } from 'socket.io-client'
import { Server, Socket as SocketForConnectedClient } from "socket.io"
import { Observable, Subject } from "rxjs";
import { createServer, request as httpRequest } from "http";
import http, { IncomingMessage, ServerResponse } from 'http';

let fromServer = new Subject<{ event: 'profile' | 'message', payload: any }>()
let toServer = new Subject<{ event: 'profile' | 'message', payload: any }>()

// startSocketServer(3001)
// startSocketServer(3002)
// startClientSocketConnection('http://localhost:3000')
startHttpServer(3001, 'http://localhost:3000/response')
consoleLog()

function consoleLog(): void {
    fromServer.subscribe(message => console.log(`From Server`, message.event, message.payload?.thisMessageID ?? 'undefined message'))
    toServer.subscribe(message => console.log(`To Server`, message.event, message.payload?.thisMessageID ?? 'undefined message'))
}

function startSocketServer(port: number): void {
    console.log(`Socket Server ${port} Started....`)
    let httpServer = createServer();
    let socketServer = new Server(httpServer)

    socketServer.on('connection', (socket: SocketForConnectedClient) => {
        socket.on(`profile`, (msg) => {
            toServer.next({ event: 'profile', payload: msg })
        })

        socket.on('message', (msg) => {
            toServer.next({ event: 'message', payload: msg })
        })

        socket.on(`disconnect`, () => {
            console.log(`I shouldn't be here lol`)
        })

        fromServer.subscribe(message => {
            socketServer.emit(message.event, message.payload)
        })
    })

    socketServer.engine.on("connection_error", (err) => {
        console.log(err.req);      // the request object
        console.log(err.code);     // the error code, for example 1
        console.log(err.message);  // the error message, for example "Session ID unknown"
        console.log(err.context);  // some additional error context
    });

    // Start the HTTP server on 127.0.0.1 with the given port
    httpServer.listen(port, '0.0.0.0', () => {
        console.log(`Socket server listening on ${port}`);
    });

}

function startClientSocketConnection(serverUrl: string): void {
    // let clientSocket = io(serverUrl)
    let clientSocket: ClientSocket = io(serverUrl, {
        reconnection: true,              // Enable automatic reconnections
        reconnectionAttempts: 1000,       // Retry up to 10 times
        reconnectionDelay: 500,          // Start with a 500ms delay
        reconnectionDelayMax: 10000,     // Delay can grow to a max of 10 seconds
        randomizationFactor: 0.3,
    })

    toServer.subscribe(message => { clientSocket.emit(message.event, message.payload) })

    clientSocket.on(`connect`, () => {

    })

    clientSocket.on(`profile`, (msg) => {
        fromServer.next({ event: 'profile', payload: msg })
    })

    clientSocket.on(`message`, (msg) => {
        fromServer.next({ event: 'message', payload: msg })
    })

    clientSocket.on(`disconnect`, () => {

    })
}


function startHttpServer(port: number, clientUrl: string): void {
    // Create an observable from the client connection
    const clientObservable = startHttpClientConnection(clientUrl);

    const server = http.createServer(async (req: IncomingMessage, res: ServerResponse) => {
        const { method, url } = req;

        if (method === 'POST' && url === '/profile') {
            // Forward the /profile request to another server (e.g., localhost:3000)
            let body = '';
            req.on('data', chunk => (body += chunk));
            req.on('end', () => {
                const options = {
                    hostname: 'localhost',
                    port: 3000,
                    path: '/profile',
                    method: 'POST',
                    headers: {
                        'Content-Type': 'application/json',
                        'Content-Length': Buffer.byteLength(body),
                    },
                };

                const forwardReq = http.request(options, forwardRes => {
                    let forwardBody = '';
                    forwardRes.on('data', chunk => (forwardBody += chunk));
                    forwardRes.on('end', () => {
                        res.writeHead(forwardRes.statusCode || 200, { 'Content-Type': 'application/json' });
                        res.end(forwardBody); // Send back the forwarded response
                    });
                });

                forwardReq.on('error', error => {
                    console.error('Error forwarding /profile request:', error);
                    res.writeHead(500, { 'Content-Type': 'application/json' });
                    res.end(JSON.stringify({ error: 'Failed to forward request' }));
                });

                forwardReq.write(body); // Send the original request body
                forwardReq.end();
            });
        } else if (method === 'POST' && url === '/response') {
            // Handle long-polling using the observable
            res.writeHead(200, { 'Content-Type': 'application/json' });

            const subscription = clientObservable.subscribe({
                next: data => {
                    res.write(JSON.stringify(data)); // Send data to the client
                    res.end(); // Close the response for this poll
                    subscription.unsubscribe(); // Unsubscribe after sending one response
                },
                error: error => {
                    console.error('Error in observable:', error);
                    res.write(JSON.stringify({ error: 'Error occurred' }));
                    res.end();
                },
            });

            // Cleanup subscription if client disconnects
            req.on('close', () => {
                subscription.unsubscribe();
            });
        } else {
            // Default 404 handler
            res.writeHead(404, { 'Content-Type': 'application/json' });
            res.end(JSON.stringify({ error: 'Not Found' }));
        }
    });

    server.listen(port, () => {
        console.log(`Server is running on http://localhost:${port}`);
    });
};


/**
 * Starts a long-polling HTTP connection to the given URL and channels responses into an Observable.
 * @param url - The URL to send the long-polling requests to.
 * @returns Observable<any> - An observable emitting the response data for each poll.
 */
function startHttpClientConnection(url: string): Observable<any> {
    return new Observable(subscriber => {
        const poll = () => {
            const options = new URL(url);

            const req = http.request(
                {
                    hostname: options.hostname,
                    port: options.port,
                    path: options.pathname + options.search,
                    method: 'GET',
                },
                res => {
                    let body = '';

                    // Accumulate data chunks
                    res.on('data', chunk => {
                        body += chunk;
                    });

                    // On response end, emit the data and start the next poll
                    res.on('end', () => {
                        try {
                            const parsedData = JSON.parse(body);
                            subscriber.next(parsedData);
                            poll(); // Start the next poll
                        } catch (error: any) {
                            subscriber.error('Error parsing response: ' + error.message);
                        }
                    });
                }
            );

            // Handle request errors
            req.on('error', error => {
                subscriber.error('Request failed: ' + error.message);
            });

            req.end(); // Send the request
        };

        poll(); // Start the polling loop

        // Cleanup logic: teardown observable if unsubscribed
        return () => {
            console.log('Stopping long-polling connection.');
        };
    });
}



// Utility function to parse JSON body
const parseBody = (req: IncomingMessage): Promise<any> => {
    return new Promise((resolve, reject) => {
        let body = '';
        req.on('data', chunk => {
            body += chunk;
        });
        req.on('end', () => {
            try {
                resolve(JSON.parse(body));
            } catch (error) {
                reject(error);
            }
        });
        req.on('error', reject);
    });
};

// Function to forward request to another server and get the response
const forwardRequest = (data: any): Promise<any> => {
    return new Promise((resolve, reject) => {
        const options = {
            hostname: 'localhost',
            port: 3000,
            path: '/profile',
            method: 'POST',
            headers: {
                'Content-Type': 'application/json',
                'Content-Length': Buffer.byteLength(JSON.stringify(data)),
            },
        };

        const req = httpRequest(options, res => {
            let responseBody = '';
            res.on('data', chunk => {
                responseBody += chunk;
            });
            res.on('end', () => {
                try {
                    resolve(JSON.parse(responseBody));
                } catch (error) {
                    reject(error);
                }
            });
        });

        req.on('error', reject);
        req.write(JSON.stringify(data));
        req.end();
    });
};