import { MessageTransmissionTransmitter } from "./msg.transmission.transmitter";
import { MessageTransmissionReceiver } from "./msg.transmission.receiver";
import { AdapterManager } from "../connector/adapter.manager";
import { EventMessage, MessageTransmissionSet, MessageTransmissionManager as MessageTransmissionManagerInterface, ReceiverProfile, TransmitterProfile } from "../interface/transport.interface";
import { v4 as uuidv4 } from 'uuid'
import { TransportEvent, Event } from "../interface/connector.interface";
import { MessageTransmissionRequestResponse } from "./msg.transmission.request-response";
import { filter, Observable, Observer, Subject } from "rxjs";
import ConsoleLogger from "../utils/log.utils";
import { TransmitterAdapter } from "../connector/adapter.transmitter";
import { ReceiverAdapter } from "../connector/adapter.receiver";
import { AdapterSet } from "../interface/general.interface";

export class MessageTransmissionManager implements MessageTransmissionManagerInterface {
    private console: ConsoleLogger = new ConsoleLogger(`MessageTransmissionManager`, ['managers'])
    private browserEnv!: boolean
    transmission: MessageTransmissionSet[] = []
    connectionManager!: AdapterManager
    event!: Subject<TransportEvent>

    constructor(browserEnv?: boolean) {
        if (browserEnv) this.browserEnv = browserEnv
        // logic here
        this.console.log({ message: `Constructing self...` })
        this.event = new Subject()
        this.connectionManager = new AdapterManager(this.event, browserEnv)

        // this.event.subscribe(event => this.console.log({ message: 'event', details: event }))

        // note that if this server is down, all these instances of transmission and connector would be lost as well. SO cannot just simply find "instances" and reuse them. Must reinstantiate them again
        this.handleEvent('Client Re-connected' as Event, this.event)
    }

    /* so there will be some changes here. will nto be assigning just one, but all of them dynamically to pour into this boy
    Transmitter only have to call this once. */
    subscribe(): Observable<MessageTransmissionSet> {
        return new Observable((observer: Observer<MessageTransmissionSet>) => {
            const targetEvent: Event = this.browserEnv ? 'New Server' : 'New Client';
            this.event.pipe(
                filter(event => event.event == targetEvent)
            ).subscribe(event => {
                // get all adapters for all the connection
                let messageTransmissionSet: MessageTransmissionSet | undefined = this.instantiateComponents((event.data as EventMessage).clientId)
                if (messageTransmissionSet) {
                    observer.next(messageTransmissionSet)
                }
            })
        })
    }

    getEvent(): Observable<TransportEvent> {
        return this.event.asObservable()
    }

    private instantiateComponents(clientId: string): MessageTransmissionSet | undefined {
        this.console.log({ message: `Instantiating new transmission set for  ${this.browserEnv ? 'Server' : 'Client'}: ${clientId}` })
        let adapterSet: AdapterSet | null = this.connectionManager.getAdapter(clientId)
        if (adapterSet) {
            let transmitter: MessageTransmissionTransmitter = this.getTransmitter(clientId, adapterSet.transmitterAdapter, this.event.asObservable())
            let receiver: MessageTransmissionReceiver = this.getReceiver(clientId, adapterSet.receiverAdapter, this.event.asObservable())
            let requestResponse: MessageTransmissionRequestResponse = this.getRequestResponse(transmitter, receiver, this.event.asObservable())
            let transmission: MessageTransmissionSet = {
                id: clientId,
                transmitter: transmitter,
                receiver: receiver,
                requestResponse: requestResponse,
                event: this.event.asObservable()
            }
            this.transmission.push(transmission)
            return transmission
        } else {
            this.console.error({ message: 'No Adapter Set' })
            return undefined
        }
    }

    private getTransmitter(transmissionId: string, adapter: TransmitterAdapter, event: Observable<TransportEvent>): MessageTransmissionTransmitter {
        let transmitterProfile: TransmitterProfile = {
            id: transmissionId,
            name: `${adapter.getInfo().transportType} Transmitter Adapter`,
            dateCreated: new Date()
        }
        return new MessageTransmissionTransmitter(transmitterProfile, adapter, event)
    }

    private getReceiver(transmissionId: string, adapter: ReceiverAdapter, event: Observable<TransportEvent>): MessageTransmissionReceiver {
        let receiverProfile: ReceiverProfile = {
            id: transmissionId,
            name: `${adapter.getInfo().transportType} Receiver Adapter`,
            dateCreated: new Date()
        }
        return new MessageTransmissionReceiver(receiverProfile, adapter, event)
    }

    private getRequestResponse(transmitterInstance: MessageTransmissionTransmitter, receiverInstance: MessageTransmissionReceiver, event: Observable<TransportEvent>): MessageTransmissionRequestResponse {
        return new MessageTransmissionRequestResponse(transmitterInstance, receiverInstance, event)
    }

    private handleEvent(eventName: Event, eventObs: Observable<TransportEvent>): void {
        eventObs.pipe(
            filter((event: TransportEvent) => event.event === eventName)
        ).subscribe(event => {
            // assuming this is reconnection case
            this.reconnectionHandler((event.data as EventMessage).clientId)
            // can include more event handlers here
        })
    }

    private reconnectionHandler(clientId: string): void {
        this.console.log({ message: `TransmissionManager: A reconnection occured. Client: ${clientId}` })
        let transmissionObj: MessageTransmissionSet | undefined = Array.from(this.transmission).find(obj => obj.id === clientId)
        if (!transmissionObj) {
            let transmissionSet: MessageTransmissionSet | undefined = this.instantiateComponents(clientId)
            if (transmissionSet) {
                this.transmission.push(transmissionSet)
            } else {
                this.console.error({ message: `Cannot find client transmission obj : ${clientId}` })
            }
        } {
            this.console.log({ message: `Transmission Object for ${clientId} Found` })
        }
    }
}


