import { Observable, Subject } from "rxjs";
import { Socket as SocketForConnectedServer } from 'socket.io-client'
import { Socket as SocketForConnectedClient } from "socket.io"
import { handleClientSocketConnection, handleNewSocketClient, startClientSocketConnection, startSocketServer } from "../utils/socket.utils";
import { ClientObject, Transport, TransportEvent, TransportMessage, TransportService } from "../interface/connector.interface";
import { WrappedMessage } from "../utils/message.ordering";
import { FisMessage } from "../interface/transport.interface";
import ConsoleLogger from "../utils/log.utils";

/* Just code in the context that this websocket service will be handling multiple UI clients. Can think about the server communication at a later time. */
export class WebsocketTransportService implements TransportService {
    private console: ConsoleLogger = new ConsoleLogger(`WebsocketTransportService`, ['transport'])
    private info: Transport = Transport.Websocket
    private connectedSocketServer: ConnectedSocketServer[] = [] // to allow the possibility of having to communicate with multiple servers as a client
    private connectedClientSocket: ConnectedSocketClient[] = [] // to keep track of the all the clients that are connected
    // private incomingMessage: Subject<TransportMessage> = new Subject() // this is only for client roles only atm
    private transportEvent!: Subject<TransportEvent>

    constructor(event: Subject<TransportEvent>) {
        this.console.log({ message: `WebsocketTransportService: Constructing socket transport service....` })
        this.transportEvent = event
        // logic here
    }

    public startServer(port: number): void {
        // logic here
        startSocketServer(port).subscribe({
            next: (connectedClient: SocketForConnectedClient) => {
                handleNewSocketClient(connectedClient, this.connectedClientSocket).subscribe({
                    next: event => this.transportEvent.next(event),
                    error: error => this.console.error({ message: `Observer Error: ${error}`, details: error }),
                    complete: () => this.console.log({ message: `Client ${connectedClient.id} disconnected...` })
                })
            },
            error: error => this.console.error({ message: `Observer Error`, details: error }),
            complete: () => this.console.log({ message: `Not sure what this is for...` })
        })
    }

    public startClient(url: string): void {
        // logic here
        startClientSocketConnection(url).then((socket: SocketForConnectedServer) => {
            handleClientSocketConnection(socket, this.connectedSocketServer).subscribe(this.transportEvent)
        }).catch((error) => {
            this.console.log({ message: `Observer Error`, details: error })
        })
    }


    public emit(message: TransportMessage): void {
        this.console.log({ message: `Emitting: ${((message.payload as WrappedMessage).payload as FisMessage).header.messageID} to ${message.target}`, details: message })
        let clientObj: ConnectedSocketClient | undefined = this.connectedClientSocket.find(obj => obj.id == message.target)
        let serverObj: ConnectedSocketServer | undefined = this.connectedSocketServer.find(obj => obj.id === message.target)
        // this.console.log({ message: `${serverObj?.connectionState.getValue(), serverObj?.id}` })
        // for server usage
        if (clientObj && clientObj.connectionState.getValue() == 'ONLINE') {
            clientObj.socketInstance.emit(`message`, message.payload)
        }
        // for client usage
        if (serverObj && serverObj.connectionState.getValue() == 'ONLINE') {
            serverObj.socketInstance.emit(`message`, message.payload)
        }
    }

    // this returns the ref pointer for the TransportEvent instantiated at Supervisor. Socket will broadcast incoming messages as event
    public subscribe(): Observable<TransportEvent> {
        return this.transportEvent.asObservable()
    }

    public getInfo(): Transport {
        return this.info
    }
}

export interface ConnectedSocketClient extends ClientObject {
    socketInstance: SocketForConnectedClient
}

export interface ConnectedSocketServer extends ClientObject {
    socketInstance: SocketForConnectedServer
}
